/*  ----------------------------------------------------------------------------

	App Title:      apeinfo
	App Version:    1.1
	Author:         Jared Breland <jbreland@legroom.net>
	Homepage:       http://www.legroom.net/software

	Program Function:
		Display information about an APE (Monkey's Audio) file

	Please visit the application's homepage for additional information.

	This application is based on the "Monkey's Audio Codec non-win32 ports"
	(MAC) project by Matthew T. Ashland.  For more information on MAC, please
	visit http://sourceforge.net/projects/mac-port/.

----------------------------------------------------------------------------  */

// Setup environment
#include <stdio.h>
#include "All.h"
#include "MACLib.h"
#include "CharacterHelper.h"
#include "APETag.h"

// Display usage information
void warning(char *prog, int retVal) {
	printf("Usage: %s [-h] [-l|-t] \"<filename.ape>\"\n", prog);
	printf("Display format and tag information about an .ape file\n");
	printf("\nOptions:\n");
	printf("   -h  Display this help information\n");
	printf("   -l  Display detailed information (default)\n");
	printf("   -t  Display tag data in VORBISCOMMENT format\n");

	// Display error message
	if (retVal == -5)
		printf("\nYou entered an unsupported option.\n");
	else if (retVal != 0)
		printf("\nError opening reading APE file. (error code %d)\n", retVal);
	exit(retVal);
}

// Convert string to uppercase
char *ucase(char *string) {
	for (unsigned int i=0; i < strlen(string); i++) {
		if ((int)string[i] >= 97 && (int)string[i] <= 122)
			string[i] = (char)(((int)string[i])-32);
	}
	return string;
}

int main(int argc, char* argv[]) {
	// Declare variables
	IAPEDecompress * pAPEDecompress = NULL;  // APE interface
	int nRetVal = 0;
	struct tag {
		tag* next;
		tag* prev;
		char* name;
		char* value;
	};
	char *pFilename = "";
	char *mode = "list";
	char *compression = "";
	unsigned int i, length, min, sec, ms;

	// Process command line arguments 
	if (argc < 2) warning(basename(argv[0]), 0);
	for (i = 1; i < argc; i++) {
		if (strcmp(argv[i],"-h") == 0) warning(basename(argv[0]), 0);
		else if (strcmp(argv[i],"-l") == 0) mode = "list";
		else if (strcmp(argv[i],"-t") == 0) mode = "tags";
		else
			if ((int)argv[i][0] == 45) warning(basename(argv[0]), -5);
	}

	// Check input file
	pFilename = argv[argc-1];
	CSmartPtr<wchar_t> spInput;
	spInput.Assign(GetUTF16FromANSI(argv[argc-1]), TRUE);
	pAPEDecompress = CreateIAPEDecompress(spInput, &nRetVal);
	if (pAPEDecompress == NULL) {
		warning(basename(argv[0]), nRetVal);
	}

	// Calculate track information
	//	Determine compression method
	switch (pAPEDecompress->GetInfo(APE_INFO_COMPRESSION_LEVEL)) {
		case COMPRESSION_LEVEL_FAST: compression = "Fast"; break;
		case COMPRESSION_LEVEL_NORMAL: compression = "Normal"; break;
		case COMPRESSION_LEVEL_HIGH: compression = "High"; break;
		case COMPRESSION_LEVEL_EXTRA_HIGH: compression = "Extra High"; break;
	}

	//	Determine tag format(s)
	CAPETag * pAPETag = (CAPETag *) pAPEDecompress->GetInfo(APE_INFO_TAG);
	BOOL bHasID3Tag = pAPETag->GetHasID3Tag();
	BOOL bHasAPETag = pAPETag->GetHasAPETag();

	//	Determine track length
	length = pAPEDecompress->GetInfo(APE_INFO_LENGTH_MS);
	ms = length % 1000;
	sec = length / 1000;
	min = sec / 60;
	sec = sec % 60;

	// Display track information 
	if (strcmp(mode, "list") == 0) {
		//	File format information
		printf("\nFile Information\n");
		printf("     File Name: %s\n", basename(pFilename));
		printf("     File Size: %d kb\n", pAPEDecompress->GetInfo(APE_INFO_APE_TOTAL_BYTES) / 1024);
		printf("   APE Version: %.2f\n", float(pAPEDecompress->GetInfo(APE_INFO_FILE_VERSION)) / float(1000));
		printf("   Compression: %s\n", compression);
		printf("    Tag Format: ");
		if (bHasID3Tag && bHasAPETag) printf("ID3, APEv%d\n", pAPETag->GetAPETagVersion() / 1000);
		else if (bHasID3Tag) printf("ID3\n");
		else if (bHasAPETag) printf("APEv%d\n", pAPETag->GetAPETagVersion() / 1000);
		else printf("None\n");

		//	Audio format information
		printf("\nAudio Information\n");
		printf("   Track Length: %d:%d.%d (%d s)\n", min, sec, ms, length/1000);
		printf("    Samples/sec: %d Hz\n", pAPEDecompress->GetInfo(APE_INFO_SAMPLE_RATE));
		printf("    Bits/sample: %d\n", pAPEDecompress->GetInfo(APE_INFO_BITS_PER_SAMPLE));
		printf("       Channels: %d\n", pAPEDecompress->GetInfo(APE_INFO_CHANNELS));
		printf("     Peak level: %d\n", pAPEDecompress->GetInfo(APE_INFO_PEAK_LEVEL));
	}

	//	Tag data
	if (bHasID3Tag || bHasAPETag) {
		if (strcmp(mode, "list") == 0) printf("\nTag Information\n");

		// Setup tag objects and declare variables
		CAPETagField * pTagField;
		tag* curr;
		tag* start = (tag*)malloc(sizeof(tag));
		curr = start;
		curr->next = (tag*)NULL;
		curr->prev = (tag*)NULL;
		unsigned int index = 0, tagnlen = 0, tagvlen = 0;
		const wchar_t *fieldName;
		char *name, *value;
		const char *fieldValue;

		// Iterate through each tag field
		while ((pTagField = pAPETag->GetTagField(index++)) != NULL) {
			fieldName = pTagField->GetFieldName();
			name = GetANSIFromUTF16(fieldName);
			fieldValue = pTagField->GetFieldValue();
			if (pAPETag->GetAPETagVersion() == CURRENT_APE_TAG_VERSION) {
				value = GetANSIFromUTF8((unsigned char *)fieldValue);
			} else {
				value = (char *)fieldValue;
			}

			// Assign tags to linked list
			curr->name = name;
			curr->value = value;
			curr->next = (tag*)malloc(sizeof(tag));
			curr->next->prev = curr;
			curr = curr->next;
			curr->next = (tag*)NULL;
			if (strlen(name) > tagnlen) tagnlen = strlen(name);
			if (strlen(value) > tagvlen) tagvlen = strlen(value);
		}

		// Sort tags list
		curr = start->next;
		tag *tester, *next;
		bool cont = true;
		while (curr->next) {
			cont = true;
			next = curr->next;
			tester = start;
			while (cont) {
				if (0 > strcmp(curr->name, tester->name)) {
					curr->prev->next = curr->next;
					curr->next->prev = curr->prev;
					curr->prev = tester->prev;
					if (curr->prev) curr->prev->next = curr;
					else start = curr;
					tester->prev = curr;
					curr->next = tester;
					cont = false;
				}
				tester = tester->next;
				if (tester->next == (tag*)NULL) cont = false;
			}
			curr=next;
		}

		// Display tags
		curr = start;
		while (curr->next != (tag*)NULL) {
			// Skip large fields like images, etc.
			if (strcmp(mode, "list") == 0) {
				for (unsigned int i = 0; i < tagnlen - strlen(curr->name) + 3; i++) printf(" ");
				if (strlen(curr->value) > 128)
					printf("%s: --- too much data to display ---\n", curr->name);
				else
					printf("%s: %s\n", curr->name, curr->value);
			} else {
				// Convert Track, Year, and Comment tag names for compatability
				if (strcmp(curr->name, "Track") == 0) {
					curr->name = (char*)malloc(12 * sizeof(char));
					strcpy(curr->name, "Tracknumber");
					if (strlen(curr->value) == 1) {
						char *temp = curr->value;
						curr->value = (char*)malloc(2 * sizeof(char));
						strcpy(curr->value, "0");
						strcat(curr->value, temp);
					}
				} else if (strcmp(curr->name, "Comment") == 0) {
					curr->name = (char*)malloc(12 * sizeof(char));
					strcpy(curr->name, "Description");
				} else if (strcmp(curr->name, "Year") == 0) {
					strcpy(curr->name, "Date");
				}
				printf("%s=%s\n", ucase(curr->name), curr->value);
			}
			curr = curr->next;
		}
	}
	
	// Cleanup objects
	delete pAPEDecompress;

	// Complete program
	return 0;
}
